package com.qmrz.config;


import com.alibaba.druid.filter.Filter;
import com.alibaba.druid.pool.DruidDataSource;
import com.alibaba.druid.wall.WallConfig;
import com.alibaba.druid.wall.WallFilter;
import com.github.pagehelper.PageInterceptor;
import com.qmrz.mybatis.MapperScannerConfigurer;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.plugin.Interceptor;
import org.apache.ibatis.session.Configuration;
import org.mybatis.spring.SqlSessionFactoryBean;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;

import javax.sql.DataSource;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

/**
 * 数据库配置工具
 */
@Slf4j
public class ConfigDBUtils {
    /**
     * 生成数据源
     *
     * @param info
     * @return
     */
    public static DataSource getDruid(ConfigDBInfo info) {
        log.info("Create DruidDataSource:" + info.getUrl());

        DruidDataSource dds = new DruidDataSource();
        dds.setUrl(info.getUrl());
        dds.setUsername(info.getUsername());
        dds.setPassword(info.getPassword());
        dds.setDriverClassName(info.getDriver());
        dds.setInitialSize(0);//初始化连接大小
        dds.setMaxActive(3000);//连接池最大使用连接数量
        dds.setMinIdle(0);//连接池最小空闲
        dds.setMaxWait(60000);//获取连接最大等待时间
        dds.setValidationQuery("select 1");//验证数据库连接有效性，要求查询语句

        WallConfig wallConfig = new WallConfig();
        wallConfig.setMultiStatementAllow(true);

        WallFilter wallFilter = new WallFilter();
        wallFilter.setConfig(wallConfig);
        List<Filter> filters = new ArrayList<>();
        filters.add(wallFilter);
        dds.setProxyFilters(filters);

        //建议配置为true，不影响性能，并且保证安全性。申请连接的时候检测，如果空闲时间大于timeBetweenEvictionRunsMillis，执行validationQuery检测连接是否有效。
        dds.setTestWhileIdle(true);

        //申请连接时执行validationQuery检测连接是否有效，配置true会降低性能。
        dds.setTestOnBorrow(false);

        //归还连接时执行validationQuery检测连接是否有效，配置true会降低性能
        dds.setTestOnReturn(false);

        //配置间隔多久才进行一次检测，检测需要关闭的空闲连接，单位是毫秒
        dds.setTimeBetweenEvictionRunsMillis(60000);

        //配置一个连接在池中最小生存的时间，单位是毫秒
        dds.setMinEvictableIdleTimeMillis(25200000);

        //对于长时间不使用的连接强制关闭
        dds.setRemoveAbandoned(true);

        //关闭超过30分钟的空闲连接，1800秒，也就是30分钟
        dds.setRemoveAbandonedTimeout(1800);

        //关闭abanded连接时输出错误日志
        dds.setLogAbandoned(true);

        //设置批量更新

        //监控数据库
        //dds.setFilters("mergeStat");
        try {
            dds.setFilters("stat,wall");
        } catch (SQLException e) {
            log.info("error:" + e.getMessage());
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return dds;
    }


    /**
     * 使用SqlSessionFactoryBean，扫描或绑定mapper(xml)位置
     *
     * @param dataSource 数据源
     * @param mapperPath mapper(xml)位置
     */
    public static SqlSessionFactoryBean getSqlSessionFactoryBean(DataSource dataSource, String mapperPath, PageInterceptor pageHelper) {
        SqlSessionFactoryBean s = new SqlSessionFactoryBean();
        s.setDataSource(dataSource);

        s.setPlugins(new Interceptor[]{pageHelper});
        ApplicationContext ctx = new FileSystemXmlApplicationContext();

        try {
            Configuration c = new Configuration();
            c.setCallSettersOnNulls(true);
            s.setConfiguration(c);
            s.setMapperLocations(ctx.getResources(mapperPath));
        } catch (IOException e) {
            e.printStackTrace();
        }

        return s;
    }

    public static PageInterceptor pageHelper() {
        PageInterceptor pageHelper = new PageInterceptor();
        Properties p = new Properties();
        p.setProperty("offsetAsPageNum", "true");
        p.setProperty("rowBoundsWithCount", "true");
        p.setProperty("reasonable", "true");
//        p.setProperty("dialect", "mysql");
        pageHelper.setProperties(p);
        return pageHelper;
    }

    /**
     * 使用MapperScannerConfigurer，扫描或绑定 mapper接口位置，并与xml文件映射生产相应实例
     *
     * @param sqlSessionFactoryBean
     * @param packageName
     */
    public static MapperScannerConfigurer getMapperScannerConfigurer(String sqlSessionFactoryBean, String packageName) {
        MapperScannerConfigurer m = new MapperScannerConfigurer();
        //包名：指定mapper接口所在包位置
        m.setBasePackage(packageName);

        //SqlSessionFactoryBea的名称，用与mapper的接口和xml的映射
        m.setSqlSessionFactoryBeanName(sqlSessionFactoryBean);

        return m;
    }
}
