package com.qmrz.mybatis;

import com.qmrz.exception.ABException;
import com.qmrz.mybatis.service.BaseService;
import com.qmrz.utils.SpringContextUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.builder.annotation.ProviderSqlSource;
import org.apache.ibatis.executor.keygen.Jdbc3KeyGenerator;
import org.apache.ibatis.executor.keygen.NoKeyGenerator;
import org.apache.ibatis.mapping.*;
import org.apache.ibatis.reflection.MetaObject;
import org.apache.ibatis.reflection.SystemMetaObject;
import org.apache.ibatis.scripting.xmltags.DynamicSqlSource;
import org.apache.ibatis.scripting.xmltags.SqlNode;
import org.apache.ibatis.scripting.xmltags.TextSqlNode;
import org.apache.ibatis.session.Configuration;
import org.apache.ibatis.session.SqlSessionFactory;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;

/**
 * mybatis功能扩展，实现无xml mapper
 */
@Slf4j
public class MapperScan {
    /**
     * 通用方法
     */
    private Map<String, Method> methods;
    private Configuration configuration;
    private BaseProviderSqlSource baseProviderSqlSource;

    public MapperScan() {
        load(SpringContextUtil.getBean(SqlSessionFactory.class));
    }

    public MapperScan(SqlSessionFactory sqlSessionFactory) {
        load(sqlSessionFactory);
    }

    public void load(SqlSessionFactory sqlSessionFactory) {
        this.methods = new HashMap<>();
        this.baseProviderSqlSource = new BaseProviderSqlSource();
        this.configuration = sqlSessionFactory.getConfiguration();

        for (Method method : BaseProviderSqlSource.class.getDeclaredMethods()) {
            methods.put(method.getName(), method);
        }
    }

    /**
     * 同时扩展mybatis接口和实体功能
     * @throws IllegalAccessException
     * @throws InvocationTargetException
     * @throws ClassNotFoundException
     */
    public void register() throws IllegalAccessException, InvocationTargetException, ClassNotFoundException {
        log.info("扩展 mybatis 接口、实体");
        registerMapper();
        registerActiveRecord();
    }

    /**
     * 实现无xml的mapper接口
     * @throws InvocationTargetException
     * @throws IllegalAccessException
     */
    public void registerMapper() throws InvocationTargetException, IllegalAccessException {
//        log.info("MapperScan registerMapper");
        for (Object object : new ArrayList<Object>(configuration.getMappedStatements())) {
            if (!(object instanceof MappedStatement)) {
                continue;
            }

            MappedStatement mappedStatement = (MappedStatement) object;
            SqlSource sqlSourceTemp = mappedStatement.getSqlSource();
            if (!(sqlSourceTemp instanceof ProviderSqlSource)) {
                continue;
            }

            log.info("mappedStatement:" + mappedStatement.getId());

            String msId = mappedStatement.getId();
            Method method = methods.get(msId.substring(msId.lastIndexOf(".") + 1));
            if (method == null) {
                log.warn("通用mapper：" + msId + " 在BaseProviderSqlSource找不到提供方法");
                continue;
            }

            Class<?> classPO = MapperUtil.getMapperClass(mappedStatement);
            SqlSource sqlSource = createSqlSource(method, mappedStatement, classPO, null);
            MetaObject msObject = SystemMetaObject.forObject(object);
            msObject.setValue("sqlSource", sqlSource);
        }
    }

    /**
     * 创建字符串类型sqlSource
     * @param script
     * @param parameterType
     * @return
     */
    private SqlSource createSqlSource(String script, Class<?> parameterType) {
        StringBuilder sql = new StringBuilder();
        sql.append("<script>\n\t");
        sql.append(script);
        sql.append("</script>");
        return configuration.getDefaultScriptingLanguageInstance()
                .createSqlSource(configuration, sql.toString(), parameterType);
    }

    /**
     * 创建sqlSource对象，自动识别sql类型（String、SqlNode）
     * @param method
     * @param mappedStatement
     * @param classPO
     * @param parameterType
     * @return
     * @throws InvocationTargetException
     * @throws IllegalAccessException
     */
    private SqlSource createSqlSource(Method method, MappedStatement mappedStatement, Class<?> classPO, Class<?> parameterType) throws InvocationTargetException, IllegalAccessException {
        SqlSource sqlSource;
        if (String.class.equals(method.getReturnType())) {
            String script = (String) method.invoke(baseProviderSqlSource, mappedStatement, classPO);
            sqlSource = createSqlSource(script, parameterType);
        } else if (SqlNode.class.isAssignableFrom(method.getReturnType())) {
            SqlNode sqlNode = (SqlNode) method.invoke(baseProviderSqlSource, mappedStatement, classPO);
            sqlSource = new DynamicSqlSource(configuration, sqlNode);
        } else {
            throw new ABException("未知sql源");
        }
        return sqlSource;
    }

    /**
     * 扩展实体功能
     * @throws ClassNotFoundException
     * @throws InvocationTargetException
     * @throws IllegalAccessException
     */
    public void registerActiveRecord() throws ClassNotFoundException, InvocationTargetException, IllegalAccessException {
        //获取实体资源
        PathMatchingResourcePatternResolver resolver = new PathMatchingResourcePatternResolver();
        Resource[] r = new Resource[0];
        try {
            r = resolver.getResources("classpath*:com/qmrz/domain/**/*.class");
        } catch (IOException e) {
            e.printStackTrace();
        }

        //实体资源转字符串
        List<String> classList = new ArrayList<>();
        try {
            for (Resource resource : r) {
                String path = resource.getURL().getPath();
                path = path.replace("\\", "/");
                classList.add(path.substring(path.indexOf("/com/") + 1
                        , path.lastIndexOf(".class"))
                        .replace("/", "."));
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

        //将属于BasePO类的实体进行扩展
        for (String str : classList) {
            Class<?> classPO = Class.forName(str);
            if (!BasePO.class.isAssignableFrom(classPO)) {
                continue;
            }
            Method[] methods = BaseService.class.getMethods();
            for (Method method : methods) {
                registerActiveRecordMethod(method.getName(), classPO);
            }
        }
    }

    /**
     * 注册实体扩展方法
     * @param methodName
     * @param classPO
     * @throws InvocationTargetException
     * @throws IllegalAccessException
     */
    private void registerActiveRecordMethod(String methodName, Class<?> classPO) throws InvocationTargetException, IllegalAccessException {
        String id = classPO.getName() + "_" + methodName;
        String newMsId = id + "_ActiveRecord";

        //创建默认动态sql
        SqlSource sqlSource = new DynamicSqlSource(configuration, new TextSqlNode("SQL_CommonMapper"));
        SqlCommandType sqlCommandType = null;

        if (methodName.startsWith("select")) {
            sqlCommandType = SqlCommandType.SELECT;
        } else if (methodName.startsWith("update")) {
            sqlCommandType = SqlCommandType.UPDATE;
        } else if (methodName.startsWith("insert")) {
            sqlCommandType = SqlCommandType.INSERT;
        } else if (methodName.startsWith("delete")) {
            sqlCommandType = SqlCommandType.DELETE;
        }

        //Statement构建配置
        MappedStatement.Builder builder =
                new MappedStatement.Builder(configuration
                        , newMsId, sqlSource, sqlCommandType);
        if (methodName.startsWith("insert")) {
            builder.keyProperty("id");
        }

        //Statement构建
        MappedStatement statement = builder.build();
        //insert后支持主键注入
        builder.keyGenerator(SqlCommandType.INSERT.equals(sqlCommandType) ? new Jdbc3KeyGenerator() : new NoKeyGenerator());

        //获取statement对应方法
        Method method = methods.get(methodName);
        //扩展方法与statement绑定并返回SqlSource
        sqlSource = createSqlSource(method, statement, classPO, null);

        //配置statement方法的sql源
        MetaObject msObject = SystemMetaObject.forObject(statement);
        msObject.setValue("sqlSource", sqlSource);

        //注入mybatis
        configuration.addMappedStatement(statement);
    }
}
