package com.qmrz.mybatis.sql.sqlnode;

import com.qmrz.exception.ABException;
import com.qmrz.mybatis.BasePO;
import com.qmrz.mybatis.SqlUtil;
import com.qmrz.mybatis.sql.BaseWrapper;
import com.qmrz.mybatis.sql.Wrapper;
import com.qmrz.mybatis.sql.WrapperUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.binding.MapperMethod;
import org.apache.ibatis.scripting.xmltags.DynamicContext;
import org.apache.ibatis.scripting.xmltags.SqlNode;

import java.util.Map;
import java.util.function.Consumer;

@Slf4j
@SuppressWarnings("unchecked")
public class SqlNodeWrapper implements SqlNode {
    /**
     * select、update、delete
     */
    private String option;
    private Class<?> classPO;
    private SqlUtil sqlUtil;

    public SqlNodeWrapper(Class<?> classPO, String option) {
        this.option = option;
        this.classPO = classPO;
        this.sqlUtil = new SqlUtil(classPO);
    }

    private void applySQL(DynamicContext context, StringBuilder sql, Map<String, Object>... param) {
        context.appendSql(sql.toString());
        for (Map<String, Object> stringObjectMap : param) {
            stringObjectMap.forEach((k, v) ->
                    context.bind(k, v)
            );
        }
    }

    @Override
    public boolean apply(DynamicContext context) {
        if (option.endsWith("Wrapper")) {
            return applyWrapper(context);//包装器模式解析
        } else {
            return applyGeneral(context);//普通模式解析
        }
    }

    /**
     * 普通模式解析
     *
     * @param context
     * @return
     */
    public boolean applyGeneral(DynamicContext context) {
        Object param = context.getBindings().get("_parameter");

        switch (option) {
            case "select":
                select(context);
                break;

            case "selectOne":
                selectOne(context);
                break;

            case "selectCount":
                selectCountResolve(context);
                break;

            case "update":
                updateResolve(context);
                break;

            case "delete":
                deleteResolve(context);
                break;

            case "insert":
                insertResolve(context);
                break;
        }
        return true;
    }

    private void select(DynamicContext context) {
        selectResolve(context, false);
    }

    private void selectOne(DynamicContext context) {
        selectResolve(context, true);
    }

    private void selectResolve(DynamicContext context, boolean isone) {
        Map<String, Object> param = (Map) context.getBindings().get("_parameter");
        StringBuilder sql = new StringBuilder();
        sql.append(sqlUtil.selectAllColumns());
        sql.append(sqlUtil.fromTable());
        sql.append(sqlUtil.whereByParam(param.keySet()));
        sql.append(sqlUtil.orderByDefault());

        if (isone) {
            sql.append(" limit 1");
        }

        applySQL(context, sql, param);
    }

    private void selectCountResolve(DynamicContext context) {
        Map<String, Object> param = (Map) context.getBindings().get("_parameter");
        StringBuilder sql = new StringBuilder();
        sql.append(sqlUtil.selectCount());
        sql.append(sqlUtil.fromTable());
        sql.append(sqlUtil.whereByParam(param.keySet()));

        applySQL(context, sql, param);
    }

    private void updateResolve(DynamicContext context) {
        UpdateParam updateParam = (UpdateParam) context.getBindings().get("_parameter");
        Map<String, Object> set = updateParam.getSetData();
        Map<String, Object> param = (Map<String, Object>) updateParam.getWhereData();

        if (param.size() == 0) {
            throw new ABException("至少有一个查询条件");
        }

        StringBuilder sql = new StringBuilder();

        sql.append(sqlUtil.updateSet(set.keySet()));
        sql.append(sqlUtil.whereByParam(param.keySet()));

        applySQL(context, sql, set, param);
    }

    private void deleteResolve(DynamicContext context) {
        Map<String, Object> param = (Map) context.getBindings().get("_parameter");
        if (param.size() == 0) {
            throw new ABException("至少有一个查询条件");
        }

        StringBuilder sql = new StringBuilder();
        sql.append(sqlUtil.deleteFrom());
        sql.append(sqlUtil.whereByParam(param.keySet()));

        applySQL(context, sql, param);
    }

    private void insertResolve(DynamicContext context) {
        Object _parameter = context.getBindings().get("_parameter");
        Map<String, Object> param;
        if (_parameter instanceof BasePO) {
//            param = new POUtil(classPO).getParamWriteNotNull(_parameter);
            param = ((BasePO)_parameter)._getProxy().getSetList();
        } else {
            param = (Map) _parameter;
        }
//        Map<String, Object> param = (Map) paramObj;
        StringBuilder sql = new StringBuilder();
        sql.append(sqlUtil.insert(param.keySet()));

        applySQL(context, sql, param);
    }

    /**
     * 包装器模式解析
     *
     * @param context
     * @return
     */
    public boolean applyWrapper(DynamicContext context) {
        log.info("SqlNodeWrapper apply");
        BaseWrapper w = null;
        switch (option) {
            case "selectWrapper":
                w = selectWrapper(context, false);
                break;
            case "selectOneWrapper":
                w = selectWrapper(context, true);
                break;
            case "updateWrapper":
                w = updateWrapper(context);
                checkWhereWrapper(w);
                break;
            case "deleteWrapper":
                w = deleteWrapper(context);
                checkWhereWrapper(w);
                break;

            case "select":
                // w = deleteWrapper(context);
                //checkWhereWrapper(w);
                break;
            default:
                throw new ABException("功能 " + option + " 未实现");
        }

        log.info(w.getSQL().toString());
        log.info(w.getSQLLimit().toString());

        log.info(String.valueOf(w.getParam()));

        //配置当前查询语句
        context.appendSql(w.getSQL().toString());
        context.appendSql(w.getSQLLimit().toString());
        //配置当前查询参数
        w.getParam().forEach((k, v) -> context.bind(k, v));

        return true;
    }

    private void checkWhereWrapper(BaseWrapper w) {
        if (!WrapperUtil.hasWhere(w)) {
            throw new ABException("必须有where条件");
        }
    }

    /**
     * 查询解析器
     *
     * @param context
     * @return
     */
    private BaseWrapper selectWrapper(DynamicContext context, boolean isone) {
        Object fn = context.getBindings().get("_parameter");
        BaseWrapper w = new Wrapper(classPO);
        BaseWrapper result = w.selectAllField().from();

        //执行表达式
        if (fn != null) {
            ((Consumer<? super BaseWrapper>) fn).accept(result);
        }
        result.orderByDefault();
        if (isone) {
            result.limit(1);
        }
        return w;
    }

    /**
     * 更新解析器
     *
     * @param context
     * @return
     */
    private BaseWrapper updateWrapper(DynamicContext context) {
        Object _parameter = context.getBindings().get("_parameter");

        Object fn;
        Map<String, Object> paramUpdateSet;
        if (_parameter instanceof UpdateParam) {
            UpdateParam up = (UpdateParam) _parameter;
            fn = up.getWhereData();
            paramUpdateSet = up.getSetData();
        } else {
            //获取update的set参数，支持BasePO实体和map
            MapperMethod.ParamMap paramMap = (MapperMethod.ParamMap) _parameter;
            Object obj = paramMap.get("param1");
            if (obj instanceof Map) {
                paramUpdateSet = (Map<String, Object>) obj;
            } else if (obj instanceof BasePO) {
                paramUpdateSet = ((BasePO) obj)._getProxy().getSetList();
//                new POUtil(obj.getClass()).getParamWriteNotNull(obj);
            } else {
                throw new ABException("参数不合法");
            }

            fn = paramMap.get("param2");
        }

        BaseWrapper w = new Wrapper(classPO);

        BaseWrapper result = w.update(paramUpdateSet);
        log.info(String.valueOf(paramUpdateSet));

        //执行表达式
        if (fn != null) {
            ((Consumer<? super BaseWrapper>) fn).accept(result);
        }

        return w;
    }

    /**
     * 删除解析器
     *
     * @param context
     * @return
     */
    private BaseWrapper deleteWrapper(DynamicContext context) {
        Object fn = context.getBindings().get("_parameter");
        BaseWrapper w = new Wrapper(classPO);
        BaseWrapper result = w.delete();

        //执行表达式
        if (fn != null) {
            ((Consumer<? super BaseWrapper>) fn).accept(result);
        }

        return w;
    }
}
