package com.qmrz.utils.tree;

import com.qmrz.mybatis.POUtil;
import com.qmrz.utils.Fn;
import com.qmrz.utils.MapChain;
import com.qmrz.utils.MapUtil;
import org.apache.commons.lang.StringUtils;
import org.apache.ibatis.reflection.MetaObject;
import org.apache.ibatis.reflection.SystemMetaObject;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * 分级信息树（如：无限级分类，省市等），构建树、扁平化、结点路径等功能
 * @param <T>
 */
public class TreeBuilder<T> {
    /**
     * 组织及扁平化后的原始数据
     */
    private List<T> flatData = new ArrayList<>();

    /**
     * 组后及扁平化后的树结点
     */
    private List<TreeNode<T>> flatTree = new ArrayList<>();

    /**
     * 初始结点（未组织过的无序列表）
     */
    private List<TreeNode<T>> treeNodes;//初始tree

    /**
     * 组织后的树
     */
    private List<TreeNode<T>> treeList;//按分级重组后

    public TreeBuilder(List<T> data) {
        this.treeNodes = new ArrayList<>();

        for (T item : data) {
            MetaObject msObject = SystemMetaObject.forObject(item);
            int id = Integer.parseInt(msObject.getValue("id").toString());
            int pid = Integer.parseInt(msObject.getValue("pid").toString());
            this.treeNodes.add(new TreeNode<T>(id, pid, item));
        }

        //构建树
        this.treeList = bulid();

        //将会初始化flatData、flatTree
        this.loadFlatData(this.treeList, 0);
    }

    /**
     * 根据List集合每一行的分类infoclassid，获得一个分类路径结果集，用与前端展示
     * 请确保List每行的实体有id和infoclassid字段
     *
     * @param list
     * @param split
     * @return Map 行id，分类路径名
     */
    public Map<String, Object> addClass(Object originData,List<?> list, String split) {
        MapChain path = MapUtil.newMap();
        for (Object item : list) {
            MetaObject mo = SystemMetaObject.forObject(item);
            Object id = mo.getValue("id");
            int infoclassid = Fn.toInt(mo.getValue("infoclassid"));
            path.put(id.toString(), pathString(infoclassid, split));
        }

        MapChain mc = MapUtil.newMap(POUtil.getMap(originData)).put("path", path.getMap());
        return mc.getMap();
    }

    /**
     * 判断结点是否在指定树（枝）里（判断 searchID 在 originID 树里是否存在）
     *
     * @param originID
     * @param searchID
     * @return
     */
    public boolean contain(int originID, int searchID) {
        List<TreeNode<T>> list = getTreeFlatList(originID);
        for (TreeNode<T> item : list) {
            if (item.getId().equals(searchID)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 获取id下结点的所有id，包括
     * @param id
     * @return
     */
    public List<Integer> treeIDList(int id) {
        List<Integer> list = new ArrayList<>();
        getTreeFlatList(id).forEach(item->list.add(item.getId()));
        return list;
    }

    /**
     * 获取id下结点数量（含id）
     * @param id
     * @return
     */
    public int treeCount(int id) {
        return getTreeFlatList(id).size();
    }

    /**
     * 根据id获取树结点
     * @param id
     * @return
     */
    public TreeNode<T> findByID(int id) {
        for (TreeNode<T> item : flatTree) {
            if (id == item.getId()) {
                return item;
            }
        }
        return null;
    }

    /**
     * 根据id获取所在路径字符串（含id）
     * @param id
     * @param split 路径结点分隔符
     * @return
     */
    public String pathString(int id, String split) {
        List<T> list = pathTree(id);
        List<String> strList = new ArrayList<>();
        for (T item : list) {
            String str = SystemMetaObject.forObject(item).getValue("name").toString();
            strList.add(str);
        }
        return StringUtils.join(strList.toArray(), split);
    }

    /**
     * 给一个id，获取路径的每个节点集合
     *
     * @param id
     * @return
     */
    public List<T> pathTree(int id) {
        List<TreeNode<T>> list = new ArrayList<>();
        pathTree(list, id);

        List<T> result = new ArrayList<>();
        for (TreeNode<T> item : list) {
            result.add(item.getItem());
        }
        Collections.reverse(result);
        return result;
    }

    /**
     * 给一个id，获取路径的每个节点集合
     *
     * @param pathList
     * @param id
     */
    private void pathTree(List<TreeNode<T>> pathList, int id) {
        TreeNode<T> node = findByID(id);
        if (node == null) {
            return;
        }
        pathList.add(node);

        for (TreeNode<T> item : flatTree) {
            if (item.getId().equals(node.getPid())) {
                pathTree(pathList, item.getId());
            }
        }
    }

    /**
     * 根据id获取此类目下（含）的所有类目的扁平化【源数据】列表
     *
     * @param id
     * @return
     */
    public List<T> getFlatDataList(int id) {
        List<T> newList = new ArrayList<>();
        List<TreeNode<T>> list = getTreeFlatList(id);
        for (TreeNode<T> item : list) {
            newList.add(item.getItem());
        }
        return newList;
    }


    /**
     * 根据id获取此类目下（含）的所有类目的扁平化【树】列表
     *
     * @param id
     * @return
     */
    public List<TreeNode<T>> getTreeFlatList(int id) {
        List<TreeNode<T>> newList = new ArrayList<>();
        getFlatList(id, false, this.treeList, newList);
        return newList;
    }

    /**
     * 根据id查询子树
     *
     * @param id
     * @param isFind  是否找到，若找到则此节点的所有子节点都加入
     * @param list
     * @param newList
     */
    private void getFlatList(int id, boolean isFind, List<TreeNode<T>> list, List<TreeNode<T>> newList) {
        for (TreeNode<T> item : list) {
            boolean tmpIsFind = false || isFind;
            if (item.getId().equals(id)) {
                tmpIsFind = true;
            }

            if (tmpIsFind) {
                newList.add(item);
            }

            if (item.getChildren() != null) {
                getFlatList(id, tmpIsFind, item.getChildren(), newList);
            }
        }
    }

    /**
     * 将重组好后的树扁平化
     *
     * @param list
     */
    private void loadFlatData(List<TreeNode<T>> list, int level) {
        for (TreeNode<T> item : list) {
            item.setLevel(level);
            SystemMetaObject.forObject(item.getItem()).setValue("level", level);
            this.flatData.add(item.getItem());
            this.flatTree.add(item);
            if (item.getChildren() != null) {
                loadFlatData(item.getChildren(), level + 1);
            }
        }
    }

    /**
     * 重组好后的树
     *
     * @return
     */
    public List<TreeNode<T>> getTreeList() {
        return treeList;
    }

    /**
     * 将重组好后的树扁平化
     *
     * @return
     */
    public List<T> getFlatData() {
        return flatData;
    }

    /**
     * 构建一颗树
     *
     * @return
     */
    private List<TreeNode<T>> bulid() {
        List<TreeNode<T>> trees = new ArrayList<>();
        for (TreeNode treeNode : treeNodes) {
            if (0 == treeNode.getPid()) {
                trees.add(treeNode);
            }

            for (TreeNode it : treeNodes) {
                if (it.getPid().equals(treeNode.getId())) {
                    if (treeNode.getChildren() == null) {
                        treeNode.setChildren(new ArrayList<>());
                    }
                    treeNode.getChildren().add(it);
                }
            }
        }
        return trees;
    }

}
